/*
 * Copyright 2013-2014 Sergey Ignatov, Alexander Zolotov
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.goide.debugger.gdb;

import com.goide.debugger.gdb.gdbmi.*;
import com.goide.debugger.gdb.messages.*;
import com.intellij.openapi.diagnostic.Logger;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.*;

/**
 * Class for interacting with GDB.
 */
public class Gdb {
  private static final Logger LOG = Logger.getInstance(Gdb.class);

  /**
   * Interface for callbacks for results from completed GDB commands.
   */
  public interface GdbEventCallback {
    /**
     * Called when a response to the command is received.
     *
     * @param event The event.
     */
    void onGdbCommandCompleted(GdbEvent event);
  }

  // Information about a command that is awaiting processing
  private static class CommandData {
    // The command
    String command;
    // The user provided callback; may be null
    GdbEventCallback callback;

    CommandData(String command, GdbEventCallback callback) {
      this.command = command;
      this.callback = callback;
    }
  }

  // Handle to the ASCII character set
  @NotNull private static final Charset ourCharset = Charset.forName("US-ASCII");

  // The listener
  private final GdbListener myListener;

  // Handle for the GDB process
  private Process myProcess;

  // Threads which read/write data from GDB
  private final Thread myReadThread;
  private Thread myWriteThread;

  // Flag indicating whether we are stopping
  @NotNull private Boolean myStopping = false;

  // Flag indicating whether we have received the first record from GDB yet
  private boolean myFirstRecord = true;

  // Token which the next GDB command will be sent with
  private long myToken = 1;

  // Commands that are waiting to be sent
  private List<CommandData> myQueuedCommands = new ArrayList<CommandData>();

  // Commands that have been sent to GDB and are awaiting a response
  private final Map<Long, CommandData> myPendingCommands = new HashMap<Long, CommandData>();

  // GDB variable objects
  private final Map<String, GdbVariableObject> myVariableObjectsByExpression = new HashMap<String, GdbVariableObject>();
  private final Map<String, GdbVariableObject> myVariableObjectsByName = new HashMap<String, GdbVariableObject>();

  // List of capabilities supported by GDB
  private Set<String> myCapabilities;

  /**
   * Constructor; prepares GDB.
   *
   * @param gdbPath          The path to the GDB executable.
   * @param workingDirectory Working directory to launch the GDB process in. May be null.
   * @param listener         Listener that is to receive GDB events.
   */
  public Gdb(final String gdbPath, final String workingDirectory, GdbListener listener) {
    // Prepare GDB
    myListener = listener;
    myReadThread = new Thread(new Runnable() {
      @Override
      public void run() {
        runGdb(gdbPath, workingDirectory);
      }
    });
  }

  /**
   * Finalizer. Kills GDB and the I/O threads.
   */
  @Override
  protected synchronized void finalize() throws Throwable {
    super.finalize();

    // Terminate the I/O threads
    if (myReadThread != null) {
      myReadThread.interrupt();
      myReadThread.join();
    }
    if (myWriteThread != null) {
      myStopping = true;
      myWriteThread.notify();
      myWriteThread.interrupt();
      myWriteThread.join();
    }

    // Kill the process
    if (myProcess != null) {
      myProcess.destroy();
    }
  }

  /**
   * Starts GDB.
   */
  public void start() {
    if (myReadThread.isAlive()) {
      throw new IllegalStateException("GDB has already been started");
    }
    myReadThread.start();
  }

  /**
   * Sends an arbitrary command to GDB.
   *
   * @param command The command to send. This may be a normal CLI command or a GDB/MI command. It
   *                should not contain any line breaks.
   */
  public void sendCommand(String command) {
    sendCommand(command, null);
  }

  /**
   * Sends an arbitrary command to GDB and requests a completion callback.
   *
   * @param command  The command to send. This may be a normal CLI command or a GDB/MI command. It
   *                 should not contain any line breaks.
   * @param callback The callback function.
   */
  public synchronized void sendCommand(String command, GdbEventCallback callback) {
    // Queue the command
    myQueuedCommands.add(new CommandData(command, callback));
    notify();
  }

  /**
   * Indicates whether GDB has the given capability.
   *
   * @param capability The capability to check for.
   * @return Whether GDB has the capability.
   */
  public synchronized boolean hasCapability(String capability) {
    if (myCapabilities == null) {
      LOG.warn("Capabilities list is null; returning 'unsupported' for capability " +
                 capability);
      return false;
    }

    return myCapabilities.contains(capability);
  }

  /**
   * Gets information about the local variables for the given stack frame.
   *
   * @param thread   The thread on which the frame resides.
   * @param frame    The frame number.
   * @param callback The callback function. This is passed a GdbVariableObjects value on success,
   *                 or GdbErrorEvent on failure.
   */
  public void getVariablesForFrame(final int thread, final int frame,
                                   @NotNull final GdbEventCallback callback) {
    // Get a list of local variables
    String command = "-stack-list-variables --thread " + thread + " --frame " + frame +
                     " --no-values";
    sendCommand(command, new GdbEventCallback() {
      @Override
      public void onGdbCommandCompleted(GdbEvent event) {
        onGdbVariablesReady(event, thread, frame, callback);
      }
    });
  }

  /**
   * Evaluates the given expression in the given context.
   *
   * @param thread     The thread to evaluate the expression in.
   * @param frame      The frame to evaluate the expression in.
   * @param expression The expression to evaluate.
   * @param callback   The callback function.
   */
  public void evaluateExpression(int thread, int frame, @NotNull final String expression,
                                 @NotNull final GdbEventCallback callback) {
    // TODO: Make this more efficient

    // Create a new variable object if necessary
    GdbVariableObject variableObject = myVariableObjectsByExpression.get(expression);
    if (variableObject == null) {
      String command = "-var-create --thread " + thread + " --frame " + frame + " - @ " +
                       GdbMiUtil.formatGdbString(expression);
      sendCommand(command, new GdbEventCallback() {
        @Override
        public void onGdbCommandCompleted(GdbEvent event) {
          onGdbNewVariableObjectReady(event, expression, callback);
        }
      });
    }

    // Update existing variable objects
    sendCommand("-var-update --thread " + thread + " --frame " + frame + " --all-values *",
                new GdbEventCallback() {
                  @Override
                  public void onGdbCommandCompleted(GdbEvent event) {
                    HashSet<String> expressions = new HashSet<String>();
                    expressions.add(expression);
                    onGdbVariableObjectsUpdated(event, expressions, callback);
                  }
                });
  }

  /**
   * Launches the GDB process and starts listening for data.
   *
   * @param gdbPath          Path to the GDB executable.
   * @param workingDirectory Working directory to launch the GDB process in. May be null.
   */
  private void runGdb(String gdbPath, @Nullable String workingDirectory) {
    try {
      // Launch the process
      final String[] commandLine = {
        gdbPath,
        "--interpreter=mi2",
        "-d",
        "$GOROOT"
      };
      File workingDirectoryFile = null;
      if (workingDirectory != null) {
        workingDirectoryFile = new File(workingDirectory);
      }
      Process process = Runtime.getRuntime().exec(commandLine, null, workingDirectoryFile);
      InputStream stream = process.getInputStream();

      // Queue startup commands
      sendCommand("-list-features", new GdbEventCallback() {
        @Override
        public void onGdbCommandCompleted(GdbEvent event) {
          onGdbCapabilitiesReady(event);
        }
      });

      // Save a reference to the process and launch the writer thread
      synchronized (this) {
        myProcess = process;
        myWriteThread = new Thread(new Runnable() {
          @Override
          public void run() {
            processWriteQueue();
          }
        });
        myWriteThread.start();
      }

      // Start listening for data
      GdbMiParser parser = new GdbMiParser();
      byte[] buffer = new byte[4096];
      int bytes;
      while ((bytes = stream.read(buffer)) != -1) {
        // Process the data
        try {
          parser.process(buffer, bytes);
        }
        catch (IllegalArgumentException ex) {
          LOG.error("GDB/MI parsing error. Current buffer contents: \"" +
                      new String(buffer, 0, bytes) + "\"", ex);
          myListener.onGdbError(ex);
          return;
        }

        // Handle the records
        List<GdbMiRecord> records = parser.getRecords();
        for (GdbMiRecord record : records) {
          handleRecord(record);
        }
        records.clear();
      }
    }
    catch (Throwable ex) {
      myListener.onGdbError(ex);
    }
  }

  /**
   * Thread function for processing the write queue.
   */
  private void processWriteQueue() {
    try {
      OutputStream stream;
      List<CommandData> queuedCommands = new ArrayList<CommandData>();
      while (true) {
        synchronized (this) {
          // Wait for some data to process
          while (myQueuedCommands.isEmpty()) {
            wait();
          }

          // Exit cleanly if we are stopping
          if (myStopping) {
            return;
          }

          // Do the processing we need before dropping out of synchronised mode
          stream = myProcess.getOutputStream();

          // Insert the commands into the pending queue
          long token = myToken;
          for (CommandData command : myQueuedCommands) {
            myPendingCommands.put(token++, command);
          }

          // Swap the queues
          List<CommandData> tmp = queuedCommands;
          queuedCommands = myQueuedCommands;
          myQueuedCommands = tmp;
        }

        // Send the queued commands to GDB
        StringBuilder sb = new StringBuilder();
        for (CommandData command : queuedCommands) {
          // Construct the message
          long token = myToken++;
          myListener.onGdbCommandSent(command.command, token);

          sb.append(token);
          sb.append(command.command);
          sb.append("\r\n");
        }
        queuedCommands.clear();

        // Send the messages
        byte[] message = sb.toString().getBytes(ourCharset);
        stream.write(message);
        stream.flush();
      }
    }
    catch (InterruptedException ex) {
      // We are exiting
    }
    catch (Throwable ex) {
      myListener.onGdbError(ex);
    }
  }

  /**
   * Handles the given GDB/MI record.
   *
   * @param record The record.
   */
  private void handleRecord(@NotNull GdbMiRecord record) {
    switch (record.type) {
      case Target:
      case Console:
      case Log:
        handleStreamRecord((GdbMiStreamRecord)record);
        break;

      case Immediate:
      case Exec:
      case Notify:
      case Status:
        handleResultRecord((GdbMiResultRecord)record);
        break;
    }

    // If this is the first record we have received we know we are fully started, so notify the
    // listener
    if (myFirstRecord) {
      myFirstRecord = false;
      myListener.onGdbStarted();
    }
  }

  /**
   * Handles the given GDB/MI stream record.
   *
   * @param record The record.
   */
  private void handleStreamRecord(GdbMiStreamRecord record) {
    // Notify the listener
    myListener.onStreamRecordReceived(record);
  }

  /**
   * Handles the given GDB/MI result record.
   *
   * @param record The record.
   */
  private void handleResultRecord(@NotNull GdbMiResultRecord record) {
    // Notify the listener
    myListener.onResultRecordReceived(record);

    // Find the pending command data
    CommandData pendingCommand = null;
    String commandType = null;
    if (record.userToken != null) {
      synchronized (this) {
        pendingCommand = myPendingCommands.remove(record.userToken);
      }
      if (pendingCommand != null) {
        // Get the command type
        int separatorIndex = pendingCommand.command.indexOf(' ');
        commandType = separatorIndex == -1 ? pendingCommand.command :
                      pendingCommand.command.substring(0, separatorIndex);
      }
    }

    // Process the event into something more useful
    GdbEvent event = GdbMiMessageConverter.processRecord(record, commandType);
    if (event != null) {
      // Notify the listener
      myListener.onGdbEventReceived(event);
      if (pendingCommand != null && pendingCommand.callback != null) {
        pendingCommand.callback.onGdbCommandCompleted(event);
      }
    }
  }

  /**
   * Callback function for when GDB has responded to our stack variables request.
   *
   * @param event    The event.
   * @param thread   The thread on which the frame resides.
   * @param frame    The frame number.
   * @param callback The user-provided callback function.
   */
  private void onGdbVariablesReady(GdbEvent event, int thread, int frame,
                                   @NotNull final GdbEventCallback callback) {
    if (event instanceof GdbErrorEvent) {
      callback.onGdbCommandCompleted(event);
      return;
    }
    if (!(event instanceof GdbVariables)) {
      GdbErrorEvent errorEvent = new GdbErrorEvent();
      errorEvent.message = "Unexpected data received from GDB";
      callback.onGdbCommandCompleted(errorEvent);
      LOG.warn("Unexpected event " + event + " received from -stack-list-variables " +
                 "request");
      return;
    }

    // Create variable objects for each of the variables if we haven't done so already
    final GdbVariables variables = (GdbVariables)event;
    for (final String variable : variables.variables.keySet()) {
      GdbVariableObject variableObject = myVariableObjectsByExpression.get(variable);
      if (variableObject == null) {
        String command = "-var-create --thread " + thread + " --frame " + frame + " - @ " +
                         variable;
        sendCommand(command, new GdbEventCallback() {
          @Override
          public void onGdbCommandCompleted(GdbEvent event) {
            onGdbNewVariableObjectReady(event, variable, callback);
          }
        });
      }
    }

    // Update any existing variable objects
    sendCommand("-var-update --thread " + thread + " --frame " + frame + " --all-values *",
                new GdbEventCallback() {
                  @Override
                  public void onGdbCommandCompleted(GdbEvent event) {
                    onGdbVariableObjectsUpdated(event, variables.variables.keySet(), callback);
                  }
                });
  }

  /**
   * Callback function for when GDB has responded to our new variable object request.
   *
   * @param event      The event.
   * @param expression The expression used to create the variable object.
   * @param callback   The user-provided callback function.
   */
  private void onGdbNewVariableObjectReady(GdbEvent event, String expression,
                                           @NotNull GdbEventCallback callback) {
    if (event instanceof GdbErrorEvent) {
      callback.onGdbCommandCompleted(event);
      return;
    }
    if (!(event instanceof GdbVariableObject)) {
      GdbErrorEvent errorEvent = new GdbErrorEvent();
      errorEvent.message = "Unexpected data received from GDB";
      callback.onGdbCommandCompleted(errorEvent);
      LOG.warn("Unexpected event " + event + " received from -var-create request");
      return;
    }

    GdbVariableObject variableObject = (GdbVariableObject)event;
    if (variableObject.name == null) {
      GdbErrorEvent errorEvent = new GdbErrorEvent();
      errorEvent.message = "Unexpected data received from GDB";
      callback.onGdbCommandCompleted(errorEvent);
      LOG.warn("Variable object returned by GDB does not have a name");
      return;
    }

    // Save the new variable object
    variableObject.expression = expression;
    myVariableObjectsByExpression.put(expression, variableObject);
    myVariableObjectsByName.put(variableObject.name, variableObject);
  }

  /**
   * Callback function for when GDB has responded to our variable objects update request.
   *
   * @param event     The event.
   * @param variables The variables the user requested.
   * @param callback  The user-provided callback function.
   */
  private void onGdbVariableObjectsUpdated(GdbEvent event, @NotNull Set<String> variables,
                                           @NotNull GdbEventCallback callback) {
    if (event instanceof GdbErrorEvent) {
      callback.onGdbCommandCompleted(event);
      return;
    }
    if (!(event instanceof GdbVariableObjectChanges)) {
      GdbErrorEvent errorEvent = new GdbErrorEvent();
      errorEvent.message = "Unexpected data received from GDB";
      callback.onGdbCommandCompleted(errorEvent);
      LOG.warn("Unexpected event " + event + " received from -var-create request");
      return;
    }

    // Update variable objects with changes
    GdbVariableObjectChanges changes = (GdbVariableObjectChanges)event;
    if (changes.changes != null) {
      for (GdbVariableObjectChange change : changes.changes) {
        if (change.name == null) {
          LOG.warn("Received a GDB variable object change with no name");
          continue;
        }

        GdbVariableObject variableObject = myVariableObjectsByName.get(change.name);
        if (variableObject == null) {
          LOG.warn("Received a GDB variable object change for a variable object " +
                     "that does not exist");
          continue;
        }

        // Set the new value
        switch (change.inScope) {
          case True:
            // Update the value
            variableObject.value = change.value;
            break;

          case False:
            // Reset the value
            variableObject.value = null;
            break;

          default:
            // TODO: Delete the variable object
            variableObject.value = null;
        }

        // Set the new type
        if (change.typeChanged && change.newType != null) {
          variableObject.type = change.newType;
        }
      }
    }

    // Construct the list of variable object the user requested
    GdbVariableObjects list = new GdbVariableObjects();
    list.objects = new ArrayList<GdbVariableObject>();
    for (String expression : variables) {
      GdbVariableObject object = myVariableObjectsByExpression.get(expression);
      if (object != null) {
        list.objects.add(object);
      }
    }

    callback.onGdbCommandCompleted(list);
  }

  /**
   * Callback function for when GDB has responded to our list capabilities request.
   *
   * @param event The event.
   */
  private void onGdbCapabilitiesReady(GdbEvent event) {
    if (event instanceof GdbErrorEvent) {
      LOG.warn("Failed to get GDB capabilities list: " + ((GdbErrorEvent)event).message);
      return;
    }
    if (!(event instanceof GdbFeatures)) {
      LOG.warn("Unexpected event " + event + " received from -list-features request");
      return;
    }

    // Save the list
    GdbFeatures features = (GdbFeatures)event;
    if (features.features != null) {
      Set<String> capabilities = new HashSet<String>(features.features);
      synchronized (this) {
        myCapabilities = capabilities;
      }
    }
  }
}
